from kmer_model import *
from base import *
from genome import reverse_complement
import random
import matplotlib.pyplot as plt
from genome import *


def log_average(a, b):
  if np.isinf(a) and np.isinf(b):
    return -np.isinf
  if a < b:
    return b + np.log(1 + np.exp(a - b)) - np.log(2)
  else:
    return a + np.log(1 + np.exp(b - a)) - np.log(2)


class model:
  def wrap_update(self, reference, read, interesting_bases, func):
    if read.strand == '-':
      reference = reverse_complement(reference)
      original_interesting_bases = interesting_bases
      interesting_bases = [reverse_complement_base(base, reference) for base in interesting_bases]

    func(reference, read, interesting_bases)

    if read.strand == '-':
      for i, base in enumerate(original_interesting_bases):
        base.sync_with_reverse_complement(interesting_bases[i])

  def update_probabilities(self, reference, read, interesting_bases):
    self.wrap_update(reference, read, interesting_bases, self.update_probabilities_internal)

  def update_probabilities_internal(self, reference, read, interesting_bases):
    pass


class window_model(model):
  def __init__(self, kmer_model = None, config = None):
    self.min_event_length = config['min_event_length']
    self.window_size = config['window_size']
    self.buffer_size = config['buffer_size']
    self.penalty = config['penalty']
    self.flashbacks = config['flashbacks']
    self.expected_SNPs = config['expected_SNPs']
    self.kmer_model = kmer_model


  def update_probabilities_c(self, reference, read, interesting_bases):
    c_kmer_model = self.kmer_model.get_c_object()
    c_read = read.get_c_object()

    numeric_reference = np.array([inv_alphabet[i] for i in reference], dtype = np.int32)
    interesting_positions = np.array([b.id for b in interesting_bases], dtype = np.int32)
    result = np.zeros(4*len(interesting_bases), dtype=np.double)
    c_wrapper.compute_probabilities(numeric_reference,
                                    len(numeric_reference),
                                    c_kmer_model,
                                    c_read,
                                    interesting_positions,
                                    len(interesting_positions),
                                    result,
                                    self.min_event_length,
                                    self.window_size // 2,
                                    self.window_size // 2,
                                    self.buffer_size,
                                    self.penalty,
                                    self.flashbacks,
                                    self.expected_SNPs)
    for i, b in enumerate(interesting_bases):
      for j in range(len(alphabet)):
        b.log_probability[j] += result[i*4 + j]

    read.free_c_object(c_read)
    self.kmer_model.free_c_object(c_kmer_model)

  def update_probabilities_internal(self, reference, read, interesting_bases):
    self.update_probabilities_c(reference, read, interesting_bases)




class moving_window_model(model):
  def __init__(self, kmer_model = None, config = None):
    self.flashbacks = config['flashbacks']
    self.min_event_length = config['min_event_length']
    self.expected_SNPs = config['expected_SNPs']
    self.kmer_model = kmer_model

  def update_probabilities_internal(self, reference, read, interesting_bases):
    c_kmer_model = self.kmer_model.get_c_object()
    c_read = read.get_c_object()

    numeric_reference = np.array([inv_alphabet[i] for i in reference], dtype=np.int32)
    result = np.zeros(4 * read.number_of_events, dtype=np.double)
    c_wrapper.compute_all_probabilities(numeric_reference,
                                    len(numeric_reference),
                                    c_kmer_model,
                                    c_read,
                                    5,
                                    result,
                                    self.min_event_length,
                                    self.flashbacks,
                                    self.expected_SNPs)
    for i, b in enumerate(interesting_bases):
      id = b.id - read.start_in_reference
      if id >= 0 and id < read.number_of_events:
        for j in range(len(alphabet)):
          b.log_probability[j] += result[id * 4 + j]

    read.free_c_object(c_read)
    self.kmer_model.free_c_object(c_kmer_model)

class no_model(model):
  def __init__(self):
    pass

  def update_base(self, base):
    for i in range(len(alphabet)):
      base.log_probability[i] -= random.random()

  def update_probabilities_internal(self, reference, read, interesting_bases):
    for base in interesting_bases:
      self.update_base(base)


def parse_cigar(cigar):
  num = 0
  result = []
  for c in cigar:
    if c >= '0' and c <= '9':
      num *= 10
      num += ord(c) - ord('0')
    else:
      result.append((num, c))
      num = 0
  return result

def aligned_length(cigar):
  return sum(num for num, c in cigar if c == 'M' or c == 'D')

def side_to_side(seq1, seq2, cigar):
  line1 = []
  line2 = []
  ind1 = 0
  ind2 = 0
  for num, c in cigar:
    if c == 'S':
      ind2 += num
    elif c == 'M':
      line1 += seq1[ind1:ind1+num]
      line2 += seq2[ind2:ind2+num]
      ind1 += num
      ind2 += num
    elif c == 'D':
      line1 += seq1[ind1:ind1 + num]
      line2 += ['-' for i in range(num)]
      ind1 += num
    elif c == 'I':
      line1 += ['-' for i in range(num)]
      line2 += seq2[ind2:ind2+num]
      ind2 += num
  print(''.join(line1[:100]))
  print(''.join(line2[:100]))

class basecall_model(model):

  def __init__(self, reference_path, config, log_filename = None):
    from bwapy import BwaAligner
    self.aligner = BwaAligner(reference_path)
    self.log_SNP_insert = np.log(config['SNP_insert'])
    self.log_non_SNP_insert = np.log(config['non_SNP_insert'])
    self.log_SNP_delete = np.log(config['SNP_delete'])
    self.log_non_SNP_delete = np.log(config['non_SNP_delete'])
    self.log_SNP_false_good = np.log(config['SNP_false_good'])
    self.log_non_SNP_good = np.log(config['non_SNP_good'])
    self.log_SNP_true_substitute = np.log(config['SNP_true_substitute'])
    self.log_SNP_false_substitute = np.log(config['SNP_false_substitute'])
    self.log_non_SNP_substitute = np.log(config['non_SNP_substitute'])
    self.log_file = None
    if log_filename:
      self.log_file = open(log_filename, 'w')

  def update_probabilities_internal(self, reference, read, interesting_bases):
    alignments = self.aligner.align_seq("".join(read.basecall))
    if len(alignments) == 0:
      return
    alignment = alignments[0]
    cigar = parse_cigar(alignment.cigar)
    aligned_reference_length = aligned_length(cigar)
    reference_aligned = []
    corresponding_read_index = [None for i in reference]
    read_aligned = []
    reference_index = alignment.pos if read.strand == '+' else len(reference) - alignment.pos - aligned_reference_length
    if read.strand == '-':
      if alignment.orient != '-':
        print("read and aligner don't agree on its orientation")
        return
      cigar = list(reversed(cigar))
    read_index = 0

    UNALIGNED, GOOD, SUBSTITUTION, DELETION, NEAR_INSERTION = 'U', 'G', 'S', 'D', 'I'
    reference_status = [UNALIGNED for i in reference[:reference_index]]

    insertions = []

    for op in cigar:
      if op[1] == 'S':
        read_index += op[0]
      elif op[1] == 'M':
        reference_aligned += reference[reference_index:reference_index+op[0]]
        for i in range(op[0]):
          corresponding_read_index[reference_index+i] = read_index+i
          if reference[reference_index+i] == read.basecall[read_index+i]:
            reference_status.append(GOOD)
          else:
            reference_status.append(SUBSTITUTION)
        reference_index += op[0]
        read_aligned += read.basecall[read_index:read_index+op[0]]
        read_index += op[0]
      elif op[1] == 'D':
        reference_aligned += reference[reference_index:reference_index + op[0]]
        reference_status += [DELETION for i in range(op[0])]
        reference_index += op[0]
        read_aligned += ['-' for i in range(op[0])]
      elif op[1] == 'I':
        read_aligned += read.basecall[read_index:read_index + op[0]]
        read_index += op[0]
        reference_aligned += ['-' for i in range(op[0])]
        insertions.append(reference_index)
      else:
        raise ValueError("Unknown operation: {}".format(op[1]))
    reference_status += [UNALIGNED for i in reference[reference_index:]]

    for idx in insertions:
      if idx > 0 and reference_status[idx-1] == GOOD:
        reference_status[idx-1] = NEAR_INSERTION
      if idx < len(reference) and reference_status[idx] == GOOD:
        reference_status[idx] = NEAR_INSERTION

    for b in interesting_bases:
      if reference_status[b.id] == UNALIGNED:
        pass
      elif reference_status[b.id] == SUBSTITUTION:
        for i, c in enumerate(alphabet):
          if c == reference[b.id]:
            b.log_probability[i] += self.log_non_SNP_substitute
          elif c == read.basecall[corresponding_read_index[b.id]]:
            b.log_probability[i] += self.log_SNP_true_substitute
          else:
            b.log_probability[i] += self.log_SNP_false_substitute
      else:
        if_SNP, if_non_SNP = None, None
        if reference_status[b.id] == GOOD:
          if_SNP = self.log_SNP_false_good
          if_non_SNP = self.log_non_SNP_good
        elif reference_status[b.id] == NEAR_INSERTION:
          if_SNP = self.log_SNP_insert
          if_non_SNP = self.log_non_SNP_insert
        elif reference_status[b.id] == DELETION:
          if_SNP = self.log_SNP_delete
          if_non_SNP = self.log_non_SNP_delete
        for i, c in enumerate(alphabet):
          if c == reference[b.id]:
            b.log_probability[i] += if_non_SNP
          else:
            b.log_probability[i] += if_SNP



      if self.log_file and reference_status[b.id] != UNALIGNED and read.strand == '+':
        self.log_file.write('{} {}'.format(b.id, reference_status[b.id]))
        if reference_status[b.id] == GOOD or reference_status[b.id] == SUBSTITUTION:
          self.log_file.write(' {}\n'.format(read.basecall[corresponding_read_index[b.id]]))
        else:
          self.log_file.write('\n')
